from dataclasses import dataclass
from hashlib import sha1, sha224, sha256, sha384, sha512
from json import load
from typing import Callable, List
from unittest import TestCase

from ..curve import P192, P224, P256, P384, P521
from ..ecdsa import sign
from ..util import RFC6979

hash_lookup = {
    '1': sha1,
    '224': sha224,
    '256': sha256,
    '384': sha384,
    '512': sha512
}


@dataclass
class TestVector:
    h: Callable[bytes, bytes]
    m: str
    k: int
    r: int
    s: int
    
    def __init__(self, data: dict):
        self.h = hash_lookup[data["h"]]
        self.m = data["m"]
        self.k = data["k"]
        self.r = data["r"]
        self.s = data["s"]


@dataclass
class TestData:
    q: int
    x: int
    vectors: List[TestVector]
        
    def __init__(self, data: dict):
        self.q = data["q"]
        self.x = data["x"]
        self.vectors = [TestVector(d) for d in data["test_case"]]


class TestRFC6979ECDSA(TestCase):
    def _run_test(self, filename, curve):
        with open(filename) as f:
            data = load(f)
        
        test_data = TestData(data)
        
        for test in test_data.vectors:
            self.assertEqual(test.k, RFC6979(test.m, test_data.x, test_data.q, test.h).gen_nonce())
            self.assertEqual((test.r, test.s), sign(test.m, test_data.x, curve=curve, hashfunc=test.h))

    def test_p192_rfc6979_ecdsa(self):
        self._run_test("../../fastecdsa/tests/vectors/rfc6979/P192.json", P192)

    def test_p224_rfc6979_ecdsa(self):
        self._run_test("../../fastecdsa/tests/vectors/rfc6979/P224.json", P224)

    def test_p256_rfc6979_ecdsa(self):
        self._run_test("../../fastecdsa/tests/vectors/rfc6979/P256.json", P256)

    def test_p384_rfc6979_ecdsa(self):
        self._run_test("../../fastecdsa/tests/vectors/rfc6979/P384.json", P384)

    def test_p521_rfc6979_ecdsa(self):
        self._run_test("../../fastecdsa/tests/vectors/rfc6979/P521.json", P521)



from . import KeyEncoder
from .util import bytes_to_int, int_bytelen, int_to_bytes
from ..curve import Curve
from ..point import Point
from ..util import mod_sqrt


class InvalidSEC1PublicKey(Exception):
    pass


class SEC1Encoder(KeyEncoder):
    binary_data = True

    @staticmethod
    def encode_public_key(point: Point, compressed: bool = True) -> bytes:
        """ Encode a public key as described in http://www.secg.org/SEC1-Ver-1.0.pdf
            in sections 2.3.3/2.3.4
                uncompressed:   04 + x_bytes + y_bytes
                compressed:     02 or 03 + x_bytes
        Args:
            point (fastecdsa.point.Point): Public key to encode
            compressed (bool): Set to False if you want an uncompressed format

        Returns:
            bytes: The SEC1 encoded public key
        """
        bytelen = int_bytelen(point.curve.q)
        if compressed:
            if point.y & 1:  # odd root
                return b'\x03' + int_to_bytes(point.x).rjust(bytelen, b'\x00')
            else:           # even root
                return b'\x02' + int_to_bytes(point.x).rjust(bytelen, b'\x00')
        return b'\x04' + int_to_bytes(point.x).rjust(bytelen, b'\x00') + int_to_bytes(point.y).rjust(bytelen, b'\x00')

    @staticmethod
    def decode_public_key(key: bytes, curve: Curve) -> Point:
        """ Decode a public key as described in http://www.secg.org/SEC1-Ver-1.0.pdf
            in sections 2.3.3/2.3.4

                uncompressed:   04 + x_bytes + y_bytes
                compressed:     02 or 03 + x_bytes

        Args:
            key (bytes): public key encoded using the SEC1 format
            curve (fastecdsa.curve.Curve): Curve to use when decoding the public key

        Returns:
            Point: The decoded public key

        Raises:
            InvalidSEC1PublicKey
        """
        bytelen = int_bytelen(curve.q)
        if key.startswith(b'\x04'):        # uncompressed key
            if len(key) != bytelen * 2 + 1:
                raise InvalidSEC1PublicKey('An uncompressed public key must be %d bytes long' % (bytelen * 2 + 1))
            x, y = bytes_to_int(key[1:bytelen + 1]), bytes_to_int(key[bytelen + 1:])
        else:                              # compressed key
            if len(key) != bytelen + 1:
                raise InvalidSEC1PublicKey('A compressed public key must be %d bytes long' % (bytelen + 1))
            x = bytes_to_int(key[1:])
            root = mod_sqrt(curve.evaluate(x), curve.p)[0]
            if key.startswith(b'\x03'):    # odd root
                y = root if root % 2 == 1 else -root % curve.p
            elif key.startswith(b'\x02'):  # even root
                y = root if root % 2 == 0 else -root % curve.p
            else:
                raise InvalidSEC1PublicKey('Wrong key format')
        return Point(x, y, curve=curve)

    @staticmethod
    def encode_private_key(d):
        raise NotImplementedError('SEC1Encoder only encodes public keys')

    @staticmethod
    def decode_private_key(data):
        raise NotImplementedError('SEC1Encoder only decodes public keys')



from unittest import TestCase

from ..curve import secp256k1
from ..encoding.pem import PEMEncoder
from ..point import Point


class TestWhitespaceParsing(TestCase):
    d = 0x3007df8e8bed6c3592a10f9d0495173dcecbc99a40e8c88b47d7d590f8b608dd
    x = 0x2c071354794f38ff439eff52f8b475b0def34a210aa3b88a63f2a7295d35f41f
    y = 0xb0f35d4f45d95657fa37d6a5431be9441a95638fc710884f43fff7938428056e
    Q = Point(x, y, curve=secp256k1)

    def test_leading_newline(self):
        keypem = "\n-----BEGIN EC PRIVATE KEY-----\n" \
                 "MHQCAQEEIDAH346L7Ww1kqEPnQSVFz3Oy8maQOjIi0fX1ZD4tgjdoAcGBSuBBAAK\n" \
                 "oUQDQgAELAcTVHlPOP9Dnv9S+LR1sN7zSiEKo7iKY/KnKV019B+w811PRdlWV/o3\n" \
                 "1qVDG+lEGpVjj8cQiE9D//eThCgFbg==\n" \
                 "-----END EC PRIVATE KEY-----\n"
        key, pubkey = PEMEncoder.decode_private_key(keypem)
        self.assertEqual(key, self.d)
        self.assertEqual(pubkey, self.Q)

    def test_leading_trailing_newlines(self):
        keypem = "\n\n\n\n-----BEGIN EC PRIVATE KEY-----\n" \
                 "MHQCAQEEIDAH346L7Ww1kqEPnQSVFz3Oy8maQOjIi0fX1ZD4tgjdoAcGBSuBBAAK\n" \
                 "oUQDQgAELAcTVHlPOP9Dnv9S+LR1sN7zSiEKo7iKY/KnKV019B+w811PRdlWV/o3\n" \
                 "1qVDG+lEGpVjj8cQiE9D//eThCgFbg==\n" \
                 "-----END EC PRIVATE KEY-----\n\n\n\n"
        key, pubkey = PEMEncoder.decode_private_key(keypem)
        self.assertEqual(key, self.d)
        self.assertEqual(pubkey, self.Q)


from binascii import unhexlify
from random import randint
from unittest import TestCase

from .. import CURVES
from fastecdsa.curve import secp256k1
from fastecdsa.ecdsa import sign
from fastecdsa.encoding.der import DEREncoder, InvalidDerSignature


class TestDEREncoder(TestCase):
    def test_encode_signature(self):
        self.assertEqual(
            DEREncoder.encode_signature(r=1, s=2),
            b"\x30"   # SEQUENCE
            b"\x06"   # Length of Sequence
            b"\x02"   # INTEGER
            b"\x01"   # Length of r
            b"\x01"   # r
            b"\x02"   # INTEGER
            b"\x01"   # Length of s
            b"\x02",  # s
        )

        # Check that we add a zero byte when the number's highest bit is set
        self.assertEqual(DEREncoder.encode_signature(r=128, s=128), b"0\x08\x02\x02\x00\x80\x02\x02\x00\x80")

        # Check a value on a standard curve like secp256k1 works
        # see https://github.com/btccom/secp256k1-go/blob/master/secp256k1/sign_vectors.yaml
        secp256k1_vectors = [
            (
                0x31a84594060e103f5a63eb742bd46cf5f5900d8406e2726dedfc61c7cf43ebad,
                unhexlify("9e5755ec2f328cc8635a55415d0e9a09c2b6f2c9b0343c945fbbfe08247a4cbe"),
                unhexlify("30440220132382ca59240c2e14ee7ff61d90fc63276325f4cbe8169fc53ade4a407c2fc802204d86fbe3bde69"
                          "75dd5a91fdc95ad6544dcdf0dab206f02224ce7e2b151bd82ab"),
            ),
            (
                0x7177f0d04c79fa0b8c91fe90c1cf1d44772d1fba6e5eb9b281a22cd3aafb51fe,
                unhexlify("2d46a712699bae19a634563d74d04cc2da497b841456da270dccb75ac2f7c4e7"),
                unhexlify("3045022100d80cf7abc9ab601373780cee3733d2cb5ff69ba1452ec2d2a058adf9645c13be0220011d1213b7d"
                          "152f72fd8759b45276ba32d9c909602e5ec89550baf3aaa8ed950"),
            ),
            (
                0x989e500d6b1397f2c5dcdf43c58ac2f14df753eb6089654e07ff946b3f84f3d5,
                unhexlify("c94f4ec84be928017cbbb447d2ab5b5d4d69e5e5fd03da7eae4378a1b1c9c402"),
                unhexlify("3045022100d0f5b740cbe3ee5b098d3c5afdefa61bb0797cb4e7b596afbd38174e1c653bb602200329e9f1a09"
                          "632de477664814791ac31544e04715db68f4b02657ba35863e711"),
            ),
            (
                0x39dfc615f2b718397f6903b0c46c47c5687e97d3d2a5e1f2b200f459f7b1219b,
                unhexlify("dfeb2092955572ce0695aa038f58df5499949e18f58785553c3e83343cd5eb93"),
                unhexlify("30440220692c01edf8aeab271df3ed4e8d57a170f014f8f9d65031aac28b5e1840acfb5602205075f9d1fdbf5"
                          "079ee052e5f3572d518b3594ef49582899ec44d065f71a55192"),
            ),
        ]

        for private_key, digest, expected in secp256k1_vectors:
            r, s = sign(digest, private_key, curve=secp256k1, prehashed=True)
            encoded = DEREncoder.encode_signature(r, s)
            self.assertEqual(encoded, expected)

    def test_decode_signature(self):
        with self.assertRaises(InvalidDerSignature):
            DEREncoder.decode_signature(b"")  # length too short
        with self.assertRaises(InvalidDerSignature):
            DEREncoder.decode_signature(b"\x31\x06\x02\x01\x01\x02\x01\x02")  # invalid SEQUENCE marker
        with self.assertRaises(InvalidDerSignature):
            DEREncoder.decode_signature(b"\x30\x07\x02\x01\x01\x02\x01\x02")  # invalid length (too short)
        with self.assertRaises(InvalidDerSignature):
            DEREncoder.decode_signature(b"\x30\x05\x02\x01\x01\x02\x01\x02")  # invalid length (too long)
        with self.assertRaises(InvalidDerSignature):
            DEREncoder.decode_signature(b"\x30\x06\x02\x03\x01\x02\x01\x02")  # invalid length of r
        with self.assertRaises(InvalidDerSignature):
            DEREncoder.decode_signature(b"\x30\x06\x02\x01\x01\x02\x03\x02")  # invalid length of s
        with self.assertRaises(InvalidDerSignature):
            DEREncoder.decode_signature(b"\x30\x06\x03\x01\x01\x02\x01\x02")  # invalid INTEGER marker for r
        with self.assertRaises(InvalidDerSignature):
            DEREncoder.decode_signature(b"\x30\x06\x02\x00\x02\x01\x02")  # length of r is 0
        with self.assertRaises(InvalidDerSignature):
            DEREncoder.decode_signature(b"\x30\x06\x02\x01\x81\x02\x01\x02")  # value of r is negative
        with self.assertRaises(InvalidDerSignature):
            DEREncoder.decode_signature(b"\x30\x07\x02\x02\x00\x01\x02\x01\x02")  # value of r starts with a zero byte
        with self.assertRaises(InvalidDerSignature):
            DEREncoder.decode_signature(b"\x30\x06\x02\x01\x01\x03\x01\x02")  # invalid INTEGER marker for s
        with self.assertRaises(InvalidDerSignature):
            DEREncoder.decode_signature(b"\x30\x06\x02\x01\x01\x02\x00")  # value of s is 0
        with self.assertRaises(InvalidDerSignature):
            DEREncoder.decode_signature(b"\x30\x06\x02\x01\x01\x02\x01\x81")  # value of s is negative
        with self.assertRaises(InvalidDerSignature):
            DEREncoder.decode_signature(b"\x30\x07\x02\x01\x01\x02\x02\x00\x02")  # value of s starts with a zero byte

        self.assertEqual(DEREncoder.decode_signature(b"\x30\x06\x02\x01\x01\x02\x01\x02"), (1, 2))
        self.assertEqual(
            DEREncoder.decode_signature(b"0\x08\x02\x02\x00\x80\x02\x02\x00\x80"), (128, 128)
        )  # verify zero bytes
        self.assertEqual(
            DEREncoder.decode_signature(b"0\x08\x02\x02\x03\xE8\x02\x02\x03\xE8"), (1000, 1000)
        )  # verify byte order

    def test_encode_decode_all_curves(self):
        for curve in CURVES:
            d = randint(1, curve.q)
            Q = d * curve.G
            r, s = sign("sign me", d, curve=curve)

            encoded = DEREncoder.encode_signature(r, s)
            decoded_r, decoded_s = DEREncoder.decode_signature(encoded)

            self.assertEqual(decoded_r, r)
            self.assertEqual(decoded_s, s)




from binascii import hexlify, unhexlify
from unittest import TestCase

from fastecdsa.curve import P256, secp192k1, secp256k1
from fastecdsa.encoding.sec1 import InvalidSEC1PublicKey, SEC1Encoder
from fastecdsa.point import Point


class TestSEC1Encoder(TestCase):
    def test_encode_public_key(self):
        # 1/ PrivateKey generated using openssl "openssl ecparam -name secp256k1 -genkey -out ec-priv.pem"
        # 2/ Printed using "openssl ec -in ec-priv.pem -text -noout" and converted to numeric using "asn1._bytes_to_int"
        priv_key = 7002880736699640265110069622773736733141182416793484574964618597954446769264
        pubkey_compressed = hexlify(SEC1Encoder.encode_public_key(secp256k1.G * priv_key))
        pubkey_uncompressed = hexlify(SEC1Encoder.encode_public_key(secp256k1.G * priv_key, compressed=False))

        # 3/ PublicKey extracted using "openssl ec -in ec-priv.pem -pubout -out ec-pub.pem"
        # 4/ Encoding verified using openssl "openssl ec -in ec-pub.pem -pubin -text -noout -conv_form compressed"
        self.assertEqual(pubkey_compressed, b"02e5e2c01985aafb6e2c3ad49f3db5ccc54b2e63343af405b521303d0f35835062")
        self.assertEqual(
            pubkey_uncompressed,
            b"04e5e2c01985aafb6e2c3ad49f3db5ccc54b2e63343af405b521303d0f3583506"
            b"23dad76df888abde5ed0cc5af1b83968edffcae5d70bedb24fdc18bb5f79499d0",
        )

        # Same with P256 Curve
        priv_P256 = 807015861248675637760562792774171551137308512372870683367415858378856470633
        pubkey_compressed = hexlify(SEC1Encoder.encode_public_key(P256.G * priv_P256))
        pubkey_uncompressed = hexlify(SEC1Encoder.encode_public_key(P256.G * priv_P256, compressed=False))
        self.assertEqual(pubkey_compressed, b"0212c9ddf64b0d1f1d91d9bd729abfb880079fa889d66604cc0b78c9cbc271824c")
        self.assertEqual(
            pubkey_uncompressed,
            b"0412c9ddf64b0d1f1d91d9bd729abfb880079fa889d66604cc0b78c9cbc271824"
            b"c9a7d581bcf2aba680b53cedbade03be62fe95869da04a168a458f369ac6a823e",
        )

        # And secp192k1 Curve
        priv_secp192k1 = 5345863567856687638748079156318679969014620278806295592453
        pubkey_compressed = hexlify(SEC1Encoder.encode_public_key(secp192k1.G * priv_secp192k1))
        pubkey_uncompressed = hexlify(SEC1Encoder.encode_public_key(secp192k1.G * priv_secp192k1, compressed=False))
        self.assertEqual(pubkey_compressed, b"03a3bec5fba6d13e51fb55bd88dd097cb9b04f827bc151d22d")
        self.assertEqual(
            pubkey_uncompressed,
            b"04a3bec5fba6d13e51fb55bd88dd097cb9b04f827bc151d22df07a73819149e8d903aa983e52ab1cff38f0d381f940d361",
        )

    def test_decode_public_key(self):
        expected_public = Point(
            x=0xE5E2C01985AAFB6E2C3AD49F3DB5CCC54B2E63343AF405B521303D0F35835062,
            y=0x3DAD76DF888ABDE5ED0CC5AF1B83968EDFFCAE5D70BEDB24FDC18BB5F79499D0,
            curve=secp256k1,
        )
        public_from_compressed = SEC1Encoder.decode_public_key(
            unhexlify(b"02e5e2c01985aafb6e2c3ad49f3db5ccc54b2e63343af405b521303d0f35835062"), secp256k1
        )
        public_from_uncompressed = SEC1Encoder.decode_public_key(
            unhexlify(
                b"04e5e2c01985aafb6e2c3ad49f3db5ccc54b2e63343af405b521303d0f3583506"
                b"23dad76df888abde5ed0cc5af1b83968edffcae5d70bedb24fdc18bb5f79499d0"
            ),
            secp256k1,
        )

        # Same values as in `test_encode_public_key`, verified using openssl
        self.assertEqual(public_from_compressed, expected_public)
        self.assertEqual(public_from_uncompressed, expected_public)
        with self.assertRaises(InvalidSEC1PublicKey) as e:
            SEC1Encoder.decode_public_key(b"\x02", secp256k1)  # invalid compressed length
        self.assertEqual(e.exception.args[0], "A compressed public key must be 33 bytes long")
        with self.assertRaises(InvalidSEC1PublicKey) as e:
            SEC1Encoder.decode_public_key(b"\x04", secp256k1)  # invalid uncompressed length
        self.assertEqual(e.exception.args[0], "An uncompressed public key must be 65 bytes long")
        with self.assertRaises(InvalidSEC1PublicKey) as e:
            # invalid prefix value
            SEC1Encoder.decode_public_key(
                unhexlify(b"05e5e2c01985aafb6e2c3ad49f3db5ccc54b2e63343af405b521303d0f35835062"), secp256k1
            )
        self.assertEqual(e.exception.args[0], "Wrong key format")

        # With P256, same values as in `test_encode_public_key`, verified using openssl
        expected_P256 = Point(
            x=0x12C9DDF64B0D1F1D91D9BD729ABFB880079FA889D66604CC0B78C9CBC271824C,
            y=0x9A7D581BCF2ABA680B53CEDBADE03BE62FE95869DA04A168A458F369AC6A823E,
            curve=P256,
        )
        public_from_compressed = SEC1Encoder.decode_public_key(
            unhexlify(b"0212c9ddf64b0d1f1d91d9bd729abfb880079fa889d66604cc0b78c9cbc271824c"), P256
        )
        self.assertEqual(public_from_compressed, expected_P256)

        # With secp192k1, same values as in `test_encode_public_key`, verified using openssl
        expected_secp192k1 = Point(
            x=0xA3BEC5FBA6D13E51FB55BD88DD097CB9B04F827BC151D22D,
            y=0xF07A73819149E8D903AA983E52AB1CFF38F0D381F940D361,
            curve=secp192k1,
        )
        public_from_compressed = SEC1Encoder.decode_public_key(
            unhexlify(b"03a3bec5fba6d13e51fb55bd88dd097cb9b04f827bc151d22d"), secp192k1
        )
        self.assertEqual(public_from_compressed, expected_secp192k1)



from unittest import TestCase

from ..keys import gen_private_key


class TestKeygen(TestCase):
    def test_gen_private_key(self):
        class FakeCurve():
            def __init__(self, q):
                self.q = q

        class FakeRandom():
            def __init__(self, values=b"\x00"):
                self.values = values
                self.pos = 0

            def __call__(self, nb):
                result = self.values[self.pos:self.pos + nb]
                self.pos += nb
                return result

        self.assertEqual(gen_private_key(FakeCurve(2), randfunc=FakeRandom(b"\x00")), 0)

        # 1 byte / 6 bits shaved off + the first try is lower than the order
        self.assertEqual(gen_private_key(FakeCurve(2), randfunc=FakeRandom(b"\x40")), 1)

        # 1 byte / 6 bits shaved off + the first try is higher than the order
        self.assertEqual(gen_private_key(FakeCurve(2), randfunc=FakeRandom(b"\xc0\x40")), 1)
        self.assertEqual(gen_private_key(FakeCurve(2), randfunc=FakeRandom(b"\xc0\x00")), 0)

        # 2 byte / 3 are shaved off, the first try is lower than the order.
        self.assertEqual(gen_private_key(FakeCurve(8191), randfunc=FakeRandom(b"\xff\xf0")), 8190)

        # 2 byte  / 3 are shaved off
        # first try : _bytes_to_int("\xff\xf8") >> 3 == 8191 (too high for order 8191)
        # second try : _bytes_to_int("\xff\xf0") >> 3 == 8190 (ok for order 8191)
        self.assertEqual(gen_private_key(FakeCurve(8191), randfunc=FakeRandom(b"\xff\xf8\xff\xf0")), 8190)

        # Same but with a different second try value
        self.assertEqual(gen_private_key(FakeCurve(8191), randfunc=FakeRandom(b"\xff\xf8\xff\xef")), 8189)


from binascii import unhexlify
from hashlib import sha224, sha256, sha384, sha3_224, sha3_256, sha3_384, sha3_512, sha512
from json import load
from unittest import TestCase, skipIf

from fastecdsa.curve import (
    P224, P256, P384, P521,
    brainpoolP224r1, brainpoolP256r1, brainpoolP320r1, brainpoolP384r1, brainpoolP512r1,
    secp256k1
)
from fastecdsa.ecdsa import verify
from fastecdsa.encoding.der import DEREncoder
from fastecdsa.encoding.sec1 import SEC1Encoder


class TestWycheproofEcdsaVerify(TestCase):
    @staticmethod
    def _get_tests(filename):
        with open(filename) as f:
            test_json = load(f)
        
        return test_json["testGroups"]

    def _test_runner(self, tests, curve, hashfunc):
        for test_group in tests:
            keybytes = unhexlify(test_group["key"]["uncompressed"])
            public_key = SEC1Encoder.decode_public_key(keybytes, curve)

            for test in test_group["tests"]:
                try:
                    message = unhexlify(test["msg"])
                    sigbytes = unhexlify(test["sig"])
                    signature = DEREncoder.decode_signature(sigbytes)
                    expected = test["result"] == "valid"

                    result = verify(signature, message, public_key, curve, hashfunc)
                    self.assertEqual(result, expected, test)
                except:
                    self.assertFalse(test["result"] == "valid", test)

    def test_brainpool224r1_sha224(self):
        filename = "../../fastecdsa/tests/vectors/whycheproof/ecdsa_brainpoolP224r1_sha224_test.json"
        tests = self._get_tests(filename)
        self._test_runner(tests, brainpoolP224r1, sha224)

    def test_brainpoolP256r1_sha256(self):
        filename = "../../fastecdsa/tests/vectors/whycheproof/ecdsa_brainpoolP256r1_sha256_test.json"
        tests = self._get_tests(filename)
        self._test_runner(tests, brainpoolP256r1, sha256)

    def test_brainpoolP320r1_sha384(self):
        filename = "../../fastecdsa/tests/vectors/whycheproof/ecdsa_brainpoolP320r1_sha384_test.json"
        tests = self._get_tests(filename)
        self._test_runner(tests, brainpoolP320r1, sha384)

    def test_brainpoolP384r1_sha384(self):
        filename = "../../fastecdsa/tests/vectors/whycheproof/ecdsa_brainpoolP384r1_sha384_test.json"
        tests = self._get_tests(filename)
        self._test_runner(tests, brainpoolP384r1, sha384)

    def test_brainpoolP512r1_sha512(self):
        filename = "../../fastecdsa/tests/vectors/whycheproof/ecdsa_brainpoolP512r1_sha512_test.json"
        tests = self._get_tests(filename)
        self._test_runner(tests, brainpoolP512r1, sha512)

    def test_p224_sha224(self):
        filename = "../../fastecdsa/tests/vectors/whycheproof/ecdsa_secp224r1_sha224_test.json"
        tests = self._get_tests(filename)
        self._test_runner(tests, P224, sha224)

    def test_p224_sha256(self):
        filename = "../../fastecdsa/tests/vectors/whycheproof/ecdsa_secp224r1_sha256_test.json"
        tests = self._get_tests(filename)
        self._test_runner(tests, P224, sha256)

    def test_p224_sha3_224(self):
        filename = "../../fastecdsa/tests/vectors/whycheproof/ecdsa_secp224r1_sha3_224_test.json"
        tests = self._get_tests(filename)
        self._test_runner(tests, P224, sha3_224)

    def test_p224_sha3_256(self):
        filename = "../../fastecdsa/tests/vectors/whycheproof/ecdsa_secp224r1_sha3_256_test.json"
        tests = self._get_tests(filename)
        self._test_runner(tests, P224, sha3_256)

    def test_p224_sha3_512(self):
        filename = "../../fastecdsa/tests/vectors/whycheproof/ecdsa_secp224r1_sha3_512_test.json"
        tests = self._get_tests(filename)
        self._test_runner(tests, P224, sha3_512)

    def test_p224_sha512(self):
        filename = "../../fastecdsa/tests/vectors/whycheproof/ecdsa_secp224r1_sha512_test.json"
        tests = self._get_tests(filename)
        self._test_runner(tests, P224, sha512)

    def test_secp256k1_sha256(self):
        filename = "../../fastecdsa/tests/vectors/whycheproof/ecdsa_secp256k1_sha256_test.json"
        tests = self._get_tests(filename)
        self._test_runner(tests, secp256k1, sha256)

    def test_secp256k1_sha3_256(self):
        filename = "../../fastecdsa/tests/vectors/whycheproof/ecdsa_secp256k1_sha3_256_test.json"
        tests = self._get_tests(filename)
        self._test_runner(tests, secp256k1, sha3_256)

    def test_secp256k1_sha3_512(self):
        filename = "../../fastecdsa/tests/vectors/whycheproof/ecdsa_secp256k1_sha3_512_test.json"
        tests = self._get_tests(filename)
        self._test_runner(tests, secp256k1, sha3_512)

    def test_secp256k1_sha512(self):
        filename = "../../fastecdsa/tests/vectors/whycheproof/ecdsa_secp256k1_sha512_test.json"
        tests = self._get_tests(filename)
        self._test_runner(tests, secp256k1, sha512)

    def test_p256_sha256(self):
        filename = "../../fastecdsa/tests/vectors/whycheproof/ecdsa_secp256r1_sha256_test.json"
        tests = self._get_tests(filename)
        self._test_runner(tests, P256, sha256)

    def test_p256_sha3_256(self):
        filename = "../../fastecdsa/tests/vectors/whycheproof/ecdsa_secp256r1_sha3_256_test.json"
        tests = self._get_tests(filename)
        self._test_runner(tests, P256, sha3_256)

    def test_p256_sha3_512(self):
        filename = "../../fastecdsa/tests/vectors/whycheproof/ecdsa_secp256r1_sha3_512_test.json"
        tests = self._get_tests(filename)
        self._test_runner(tests, P256, sha3_512)

    def test_p256_sha512(self):
        filename = "../../fastecdsa/tests/vectors/whycheproof/ecdsa_secp256r1_sha512_test.json"
        tests = self._get_tests(filename)
        self._test_runner(tests, P256, sha512)

    def test_p384_sha384(self):
        filename = "../../fastecdsa/tests/vectors/whycheproof/ecdsa_secp384r1_sha384_test.json"
        tests = self._get_tests(filename)
        self._test_runner(tests, P384, sha384)

    def test_p384_sha3_384(self):
        filename = "../../fastecdsa/tests/vectors/whycheproof/ecdsa_secp384r1_sha3_384_test.json"
        tests = self._get_tests(filename)
        self._test_runner(tests, P384, sha3_384)

    def test_p384_sha3_512(self):
        filename = "../../fastecdsa/tests/vectors/whycheproof/ecdsa_secp384r1_sha3_512_test.json"
        tests = self._get_tests(filename)
        self._test_runner(tests, P384, sha3_512)

    def test_p384_sha512(self):
        filename = "../../fastecdsa/tests/vectors/whycheproof/ecdsa_secp384r1_sha512_test.json"
        tests = self._get_tests(filename)
        self._test_runner(tests, P384, sha512)

    def test_p521_sha3_512(self):
        filename = "../../fastecdsa/tests/vectors/whycheproof/ecdsa_secp521r1_sha3_512_test.json"
        tests = self._get_tests(filename)
        self._test_runner(tests, P521, sha3_512)

    def test_p521_sha512(self):
        filename = "../../fastecdsa/tests/vectors/whycheproof/ecdsa_secp521r1_sha512_test.json"
        tests = self._get_tests(filename)
        self._test_runner(tests, P521, sha512)



from binascii import b2a_uu
from os import urandom
from timeit import timeit

from .curve import (
    Curve, P192, P224, P256, P384, P521, W25519, W448, secp192k1, secp224k1, secp256k1,
    brainpoolP160r1, brainpoolP192r1, brainpoolP224r1, brainpoolP256r1, brainpoolP320r1,
    brainpoolP384r1, brainpoolP512r1
)
from .ecdsa import sign, verify
from .keys import gen_keypair
from .point import Point


def sign_and_verify(d: int, Q: Point, curve: Curve):
    msg = b2a_uu(urandom(32))
    sig = sign(msg, d, curve=curve)
    assert verify(sig, msg, Q, curve=curve)


if __name__ == '__main__':
    iterations = 1000
    curves = (
        P192, P224, P256, P384, P521, W25519, W448, secp192k1, secp224k1, secp256k1,
        brainpoolP160r1, brainpoolP192r1, brainpoolP224r1, brainpoolP256r1, brainpoolP320r1,
        brainpoolP384r1, brainpoolP512r1
    )

    for curve in curves:
        d, Q = gen_keypair(curve)
        time = timeit(stmt=lambda: sign_and_verify(d, Q, curve), number=iterations)
        print('{} signatures and verifications with curve {} took {:.2f} seconds'.format(
            iterations, curve.name, time))



from struct import pack, unpack

from ..curve import Curve
from ..point import Point
from .util import int_bytelen, int_to_bytes

INTEGER = b'\x02'
BIT_STRING = b'\x03'
OCTET_STRING = b'\x04'
OBJECT_IDENTIFIER = b'\x06'
SEQUENCE = b'\x30'
PARAMETERS = b'\xa0'
PUBLIC_KEY = b'\xa1'


class ASN1EncodingError(Exception):
    pass


def _asn1_len(data: bytes) -> bytes:
    # https://www.itu.int/ITU-T/studygroups/com17/languages/X.690-0207.pdf
    # section 8.1.3.3
    dlen = len(data)

    if dlen < 0x80:
        return pack('=B', dlen)
    else:
        encoded = b''

        while dlen:
            len_byte = pack('=B', dlen & 0xff)
            encoded = len_byte + encoded
            dlen >>= 8

        return pack('=B', 0x80 | len(encoded)) + encoded


def asn1_structure(data_type: bytes, data: bytes) -> bytes:
    return data_type + _asn1_len(data) + data


def asn1_private_key(d: int, curve: Curve) -> bytes:
    d_bytes = int_to_bytes(d)
    padding = b'\x00' * (int_bytelen(curve.q) - len(d_bytes))
    return asn1_structure(OCTET_STRING, padding + d_bytes)


def asn1_ecversion(version=1) -> bytes:
    version_bytes = int_to_bytes(version)
    return asn1_structure(INTEGER, version_bytes)


def asn1_ecpublickey() -> bytes:
    # via RFC 5480 - The "unrestricted" algorithm identifier is:
    # id-ecPublicKey OBJECT IDENTIFIER ::= {
    #   iso(1) member-body(2) us(840) ansi-X9-62(10045) keyType(2) 1 }
    return asn1_structure(OBJECT_IDENTIFIER, b'\x2A\x86\x48\xCE\x3D\x02\x01')


def asn1_oid(curve: Curve) -> bytes:
    oid_bytes = curve.oid
    return asn1_structure(OBJECT_IDENTIFIER, oid_bytes)


def asn1_public_key(Q: Point) -> bytes:
    p_len = int_bytelen(Q.curve.p)

    x_bytes = int_to_bytes(Q.x)
    x_padding = b'\x00' * (p_len - len(x_bytes))

    y_bytes = int_to_bytes(Q.y)
    y_padding = b'\x00' * (p_len - len(y_bytes))

    key_bytes = b'\x00\x04' + x_padding + x_bytes + y_padding + y_bytes
    return asn1_structure(BIT_STRING, key_bytes)


def parse_asn1_length(data: bytes) -> (int, bytes, bytes):
    """
    Parse an ASN.1 encoded structure.

    Args:
        data (bytes): A sequence of bytes representing an ASN.1 encoded structure

    Returns:
        (int, bytes, bytes): A tuple of the integer length in bytes, the byte representation of the integer,
                             and the remaining bytes after the integer bytes in the sequence
    """
    (initial_byte,) = unpack('=B', data[:1])
    data = data[1:]

    if not (initial_byte & 0x80):
        length = initial_byte
    else:
        count = initial_byte & 0x7f
        fmt = {1: '=B', 2: '=H', 3: '=L', 4: '=L', 5: '=Q', 6: '=Q', 7: '=Q', 8: '=Q'}
        zero_padding = b'\x00' * ((1 << (count.bit_length() - 1)) - count)

        (length,) = unpack(fmt[count], zero_padding + data[:count])
        data = data[count:]

    if length > len(data):
        raise ASN1EncodingError("Parsed length of ASN.1 structure to be {} bytes but only {} bytes"
                                "remain in the provided data".format(length, len(data)))

    return length, data[:length], data[length:]


def parse_asn1_int(data: bytes) -> (int, bytes, bytes):
    """
    Parse an ASN.1 encoded integer.

    Args:
        data (bytes): A sequence of bytes whose start is an ASN.1 integer encoding

    Returns:
        (int, bytes, bytes): A tuple of the integer length in bytes, the byte representation of the integer,
                             and the remaining bytes after the integer bytes in the sequence
    """

    # encoding needs at least the type, length and data
    if len(data) < 3:
        raise ASN1EncodingError("ASN.1 encoded integer must be at least 3 bytes long")
    # integer should be identified as ASN.1 integer
    if data[0] != ord(INTEGER):
        raise ASN1EncodingError("Value should be a ASN.1 INTEGER")

    length, data, remaining = parse_asn1_length(data[1:])

    # integer length should match length indicated
    if length != len(data):
        raise ASN1EncodingError("Expected ASN.1 INTEGER to be {} bytes, got {} bytes".format(length, len(data)))

    return length, data, remaining



from binascii import hexlify
from hashlib import sha256
from typing import Tuple, TypeVar

from fastecdsa import _ecdsa
from .curve import Curve, P256
from .point import Point
from .util import RFC6979, msg_bytes

MsgTypes = TypeVar('MsgTypes', str, bytes, bytearray)
SigType = Tuple[int, int]


class EcdsaError(Exception):
    def __init__(self, msg):
        self.msg = msg


def sign(msg: MsgTypes, d: int, curve: Curve = P256, hashfunc=sha256, prehashed: bool = False):
    """Sign a message using the elliptic curve digital signature algorithm.

    The elliptic curve signature algorithm is described in full in FIPS 186-4 Section 6. Please
    refer to http://nvlpubs.nist.gov/nistpubs/FIPS/NIST.FIPS.186-4.pdf for more information.

    Args:
        |  msg (str|bytes|bytearray): A message to be signed.
        |  d (int): The ECDSA private key of the signer.
        |  curve (fastecdsa.curve.Curve): The curve to be used to sign the message.
        |  hashfunc (_hashlib.HASH): The hash function used to compress the message.
        |  prehashed (bool): The message being passed has already been hashed by :code:`hashfunc`.

    Returns:
        (int, int): The signature (r, s) as a tuple.
    """
    # generate a deterministic nonce per RFC6979
    rfc6979 = RFC6979(msg, d, curve.q, hashfunc, prehashed=prehashed)
    k = rfc6979.gen_nonce()

    # Fix the bit-length of the random nonce,
    # so that it doesn't leak via timing.
    # This does not change that ks (mod n) = kt (mod n) = k (mod n)
    ks = k + curve.q
    kt = ks + curve.q
    if ks.bit_length() == curve.q.bit_length():
        k = kt
    else:
        k = ks

    if prehashed:
        hex_digest = hexlify(msg).decode()
    else:
        hex_digest = hashfunc(msg_bytes(msg)).hexdigest()

    r, s = _ecdsa.sign(
        hex_digest,
        str(d),
        str(k),
        str(curve.p),
        str(curve.a),
        str(curve.b),
        str(curve.q),
        str(curve.gx),
        str(curve.gy)
    )
    return int(r), int(s)


def verify(
        sig: SigType, msg: MsgTypes, Q: Point, curve: Curve = P256, hashfunc=sha256, prehashed: bool = False) -> bool:
    """Verify a message signature using the elliptic curve digital signature algorithm.

    The elliptic curve signature algorithm is described in full in FIPS 186-4 Section 6. Please
    refer to http://nvlpubs.nist.gov/nistpubs/FIPS/NIST.FIPS.186-4.pdf for more information.

    Args:
        |  sig (int, int): The signature for the message.
        |  msg (str|bytes|bytearray): A message to be signed.
        |  Q (fastecdsa.point.Point): The ECDSA public key of the signer.
        |  curve (fastecdsa.curve.Curve): The curve to be used to sign the message.
        |  hashfunc (_hashlib.HASH): The hash function used to compress the message.
        |  prehashed (bool): The message being passed has already been hashed by :code:`hashfunc`.

    Returns:
        bool: True if the signature is valid, False otherwise.

    Raises:
        fastecdsa.ecdsa.EcdsaError: If the signature or public key are invalid. Invalid signature
            in this case means that it has values less than 1 or greater than the curve order.
    """
    if isinstance(Q, tuple):
        Q = Point(Q[0], Q[1], curve)
    r, s = sig

    # validate Q, r, s (Q should be validated in constructor of Point already but double check)
    if not curve.is_point_on_curve((Q.x, Q.y)):
        raise EcdsaError('Invalid public key, point is not on curve {}'.format(curve.name))
    elif r > curve.q or r < 1:
        raise EcdsaError(
            'Invalid Signature: r is not a positive integer smaller than the curve order')
    elif s > curve.q or s < 1:
        raise EcdsaError(
            'Invalid Signature: s is not a positive integer smaller than the curve order')

    if prehashed:
        hashed = hexlify(msg).decode()
    else:
        hashed = hashfunc(msg_bytes(msg)).hexdigest()

    return _ecdsa.verify(
        str(r),
        str(s),
        hashed,
        str(Q.x),
        str(Q.y),
        str(curve.p),
        str(curve.a),
        str(curve.b),
        str(curve.q),
        str(curve.gx),
        str(curve.gy)
    )



from hashlib import sha1, sha224, sha256, sha384, sha512
from unittest import TestCase

from ..util import RFC6979


class TestNonceGeneration(TestCase):
    def test_rfc_6979(self):
        msg = 'sample'
        x = 0x09A4D6792295A7F730FC3F2B49CBC0F62E862272F
        q = 0x4000000000000000000020108A2E0CC0D99F8A5EF

        expected = 0x09744429FA741D12DE2BE8316E35E84DB9E5DF1CD
        nonce = RFC6979(msg, x, q, sha1).gen_nonce()
        self.assertTrue(nonce == expected)

        expected = 0x323E7B28BFD64E6082F5B12110AA87BC0D6A6E159
        nonce = RFC6979(msg, x, q, sha224).gen_nonce()
        self.assertTrue(nonce == expected)

        expected = 0x23AF4074C90A02B3FE61D286D5C87F425E6BDD81B
        nonce = RFC6979(msg, x, q, sha256).gen_nonce()
        self.assertTrue(nonce == expected)

        expected = 0x2132ABE0ED518487D3E4FA7FD24F8BED1F29CCFCE
        nonce = RFC6979(msg, x, q, sha384).gen_nonce()
        self.assertTrue(nonce == expected)

        expected = 0x00BBCC2F39939388FDFE841892537EC7B1FF33AA3
        nonce = RFC6979(msg, x, q, sha512).gen_nonce()
        self.assertTrue(nonce == expected)


from unittest import TestCase

from ..curve import W25519
from ..point import Point


class TestTypeValidation(TestCase):
    def test_type_validation_add(self):
        with self.assertRaises(TypeError):
            _ = Point.IDENTITY_ELEMENT + 2

        with self.assertRaises(TypeError):
            _ = W25519.G + 2

        with self.assertRaises(TypeError):
            _ = 2 + Point.IDENTITY_ELEMENT

        with self.assertRaises(TypeError):
            _ = 2 + W25519.G

    def test_type_validation_sub(self):
        with self.assertRaises(TypeError):
            _ = Point.IDENTITY_ELEMENT - 2

        with self.assertRaises(TypeError):
            _ = W25519.G - 2

        with self.assertRaises(TypeError):
            _ = 2 - Point.IDENTITY_ELEMENT

        with self.assertRaises(TypeError):
            _ = 2 - W25519.G

    def test_type_validation_mul(self):
        with self.assertRaises(TypeError):
            _ = Point.IDENTITY_ELEMENT * 1.5

        with self.assertRaises(TypeError):
            _ = W25519.G * 1.5

        with self.assertRaises(TypeError):
            _ = 1.5 * Point.IDENTITY_ELEMENT

        with self.assertRaises(TypeError):
            _ = 1.5 * W25519.G

    def test_point_at_infinity(self):
        self.assertEqual(W25519.G * 0, Point.IDENTITY_ELEMENT)
        self.assertEqual(W25519.G * W25519.q, Point.IDENTITY_ELEMENT)
        self.assertEqual(W25519.G + Point.IDENTITY_ELEMENT, W25519.G)
        self.assertEqual(W25519.G - Point.IDENTITY_ELEMENT, W25519.G)



from fastecdsa import curvemath
from .curve import P256
from .util import validate_type


class CurveMismatchError(Exception):
    def __init__(self, curve1, curve2):
        self.msg = 'Tried to add points on two different curves <{}> & <{}>'.format(
            curve1.name, curve2.name)


class Point:
    """Representation of a point on an elliptic curve.

    Attributes:
        |  x (int): The x coordinate of the point.
        |  y (int): The y coordinate of the point.
        |  curve (:class:`Curve`): The curve that the point lies on.
    """

    def __init__(self, x: int, y: int, curve=P256):
        r"""Initialize a point on an elliptic curve.

        The x and y parameters must satisfy the equation :math:`y^2 \equiv x^3 + ax + b \pmod{p}`,
        where a, b, and p are attributes of the curve parameter.

        Args:
            |  x (int): The x coordinate of the point.
            |  y (int): The y coordinate of the point.
            |  curve (:class:`Curve`): The curve that the point lies on.
        """

        # Reduce numbers before computation to avoid errors and limit computations.
        if curve is not None:
            x = x % curve.p
            y = y % curve.p

        if not (x == 0 and y == 0 and curve is None) and not curve.is_point_on_curve((x, y)):
            raise ValueError(
                'coordinates are not on curve <{}>\n\tx={:x}\n\ty={:x}'.format(curve.name, x, y))
        else:
            self.x = x
            self.y = y
            self.curve = curve

    def __str__(self):
        if self == self.IDENTITY_ELEMENT:
            return '<POINT AT INFINITY>'
        else:
            return 'X: 0x{:x}\nY: 0x{:x}\n(On curve <{}>)'.format(self.x, self.y, self.curve.name)

    def __unicode__(self) -> str:
        return self.__str__()

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other) -> bool:
        validate_type(other, Point)
        return self.x == other.x and self.y == other.y and self.curve is other.curve

    def __add__(self, other):
        """Add two points on the same elliptic curve.

        Args:
            | self (:class:`Point`): a point :math:`P` on the curve
            | other (:class:`Point`): a point :math:`Q` on the curve

        Returns:
            :class:`Point`: A point :math:`R` such that :math:`R = P + Q`
        """
        validate_type(other, Point)

        if self == self.IDENTITY_ELEMENT:
            return other
        elif other == self.IDENTITY_ELEMENT:
            return self
        elif self.curve is not other.curve:
            raise CurveMismatchError(self.curve, other.curve)
        elif self == -other:
            return self.IDENTITY_ELEMENT
        else:
            x, y = curvemath.add(
                str(self.x),
                str(self.y),
                str(other.x),
                str(other.y),
                str(self.curve.p),
                str(self.curve.a),
                str(self.curve.b),
                str(self.curve.q),
                str(self.curve.gx),
                str(self.curve.gy)
            )
            return Point(int(x), int(y), self.curve)

    def __radd__(self, other):
        """Add two points on the same elliptic curve.

        Args:
            | self (:class:`Point`): a point :math:`P` on the curve
            | other (:class:`Point`): a point :math:`Q` on the curve

        Returns:
            :class:`Point`: A point :math:`R` such that :math:`R = R + Q`
        """
        validate_type(other, Point)
        return self.__add__(other)

    def __sub__(self, other):
        """Subtract two points on the same elliptic curve.

        Args:
            | self (:class:`Point`): a point :math:`P` on the curve
            | other (:class:`Point`): a point :math:`Q` on the curve

        Returns:
            :class:`Point`: A point :math:`R` such that :math:`R = P - Q`
        """
        validate_type(other, Point)

        if self == other:
            return self.IDENTITY_ELEMENT
        elif other == self.IDENTITY_ELEMENT:
            return self

        negative = Point(other.x, -other.y % other.curve.p, other.curve)
        return self.__add__(negative)

    def __mul__(self, scalar: int):
        r"""Multiply a :class:`Point` on an elliptic curve by an integer.

        Args:
            | self (:class:`Point`): a point :math:`P` on the curve
            | other (int): an integer :math:`d \in \mathbb{Z_q}` where :math:`q` is the order of
                the curve that :math:`P` is on

        Returns:
            :class:`Point`: A point :math:`R` such that :math:`R = P * d`
        """
        validate_type(scalar, int)

        if scalar == 0:
            return self.IDENTITY_ELEMENT

        x, y = curvemath.mul(
            str(self.x),
            str(self.y),
            str(abs(scalar)),
            str(self.curve.p),
            str(self.curve.a),
            str(self.curve.b),
            str(self.curve.q),
            str(self.curve.gx),
            str(self.curve.gy)
        )
        x = int(x)
        y = int(y)
        if x == 0 and y == 0:
            return self.IDENTITY_ELEMENT
        return Point(x, y, self.curve) if scalar > 0 else -Point(x, y, self.curve)

    def __rmul__(self, scalar: int):
        r"""Multiply a :class:`Point` on an elliptic curve by an integer.

        Args:
            | self (:class:`Point`): a point :math:`P` on the curve
            | other (long): an integer :math:`d \in \mathbb{Z_q}` where :math:`q` is the order of
                the curve that :math:`P` is on

        Returns:
            :class:`Point`: A point :math:`R` such that :math:`R = d * P`
        """
        return self.__mul__(scalar)

    def __neg__(self):
        """Return the negation of a :class:`Point` i.e. the points reflection over the x-axis.

        Args:
            | self (:class:`Point`): a point :math:`P` on the curve

        Returns:
            :class:`Point`: A point :math:`R = (P_x, -P_y)`
        """
        if self == self.IDENTITY_ELEMENT:
            return self

        return Point(self.x, -self.y % self.curve.p, self.curve)


Point.IDENTITY_ELEMENT = Point(0, 0, curve=None)  # also known as the point at infinity
